from . import *
from .Authenticator import Authenticator
import time
import io
import requests
import logging
from urllib3.util.retry import Retry

from .constants import USER_AGENT

log = logging.getLogger(LOGGER_NAME)

class HTTPCaller:
    def __init__(self):
        self.url = config.endpoint
        self.jobid = config.job_id
        self.eid = config.eid
        self.token = Authenticator().get_token()
        self.user_agent = USER_AGENT

    def set_headers(self, headers=None):
        if headers is None:
            headers = {}
        # adding token to the headers
        headers.update(self.token)
        headers.update({'X-KH-JOB-ID': self.jobid})
        headers.update({'User-Agent': self.user_agent})
        return headers

    def post(self, path, payload=None, data=None, headers=None):
        headers = self.set_headers(headers)

        log.debug(f'Sending POST request to {path}, with payload: {payload}')
        r = requests.post(self.url + path, json=payload, data=data, headers=headers)
        if (r.status_code >= 200) and (r.status_code < 300):
            log.info('Posting with message {} was successful with response: {}'.format(payload, r.text))
            return r

        else:
            log.error('Posting with message {} ended with error (error code: {}). Response: {}'.format(payload, r.status_code, r.text))
            return r

    def get(self, path, headers=None):
        headers = self.set_headers(headers)

        # Request URL
        full_path = self.url + path
        log.debug('GET request before sending: {}, {}'.format(full_path, headers))

        ### Send GET request

        # Create a session
        s = requests.Session()

        # Define your retries for http and https urls
        http_retries = Retry(backoff_factor=0.1, total=5, status_forcelist = [104, 413, 429, 503, 502, 500])
        https_retries = Retry(backoff_factor=0.1, total=5, status_forcelist = [104, 413, 429, 503, 502, 500])

        # Create adapters with the retry logic for each
        http = requests.adapters.HTTPAdapter(max_retries=http_retries)
        https = requests.adapters.HTTPAdapter(max_retries=https_retries)

        # Replace the session’s original adapters
        s.mount('http://', http)
        s.mount('https://', https)

        r = s.get(full_path, headers=headers)

        log.debug(f'GET: {r.url}')
        log.debug(f'Headers: {r.request.headers}')

        log.debug('>> GET method returned a {} status'.format(r.status_code))

        if (r.status_code >= 200) and (r.status_code < 300):
            log.debug('Data received from {} with response: {}'.format(path, r.text))
            return r

        else:
            log.error('Data reception from {} ended with error (error code: {}). Response: {}'.format(path, r.status_code, r.text))
            return r

    def file_uploader(self, payload, file, headers=None):
        '''
        Uploads a file from disk or an object from RAM to datalake with a PUT function
        :param payload:
        :param file: python object to stream or path to a disk file to upload
        :param headers:
        :return:
        '''

        headers = self.set_headers(headers)

        # send meta data
        file_params_json = self.post('/file', payload, headers=headers).json()
        log.debug(file_params_json)
        url = file_params_json['location']
        file_id = file_params_json['fileId']

        # Get the file content to upload as binary stream
        if isinstance(file, str):
            file_content = open(file, 'rb').read()

        elif isinstance(file, io.BytesIO):
            # Stream the python object directly
            file_content = file

        else:
            raise TypeError(
                'Passed file argument must be either a file path to a file saved on the disk, '
                'either a io.BytesIO object to stream from memory')

        # PUT request
        r = requests.put(url,
                         data=file_content,
                         headers={'Content-Type': 'application/octet-stream', "User-Agent": self.user_agent}
                         )

        log.debug(f'PUT {r.url} \nHEADERS: {r.request.headers}')

        # Returned code after uploading
        if (r.status_code >= 200) and (r.status_code < 300):

            # wait and ping every 10 seconds where the upload is finished - up to 10 times
            i = 0
            while i < 10:
                r_sync_ = self.get('/file/search?size=1&from=0&query=uuid%3D{}'.format(file_id), headers=headers)

                r_sync = r_sync_.json()

                i += 1
                log.debug(f'Search the uploaded file for the {i} time')

                # if nothing in items than wait 10 seconds
                if r_sync['items'] == []:
                    # if it is 10hth wait, something is wrong and exception is thrown
                    if i == 10:
                        raise Exception(f"File {file_id} upload is not successful after 100 seconds of waiting")
                    time.sleep(10)
                # otherwise items are not empty and file is successfully uploaded
                else:
                    log.debug(r_sync['items'])
                    log.info('File uploaded')
                    break

        return file_id

    def get_from_S3(self, S3path):

        # Create a session
        s = requests.Session()

        # Define your retries for http and https urls
        http_retries = Retry(backoff_factor=0.1, total=5, status_forcelist = [413, 429, 503, 502, 500])
        https_retries = Retry(backoff_factor=0.1, total=5, status_forcelist = [413, 429, 503, 502, 500])

        # Create adapters with the retry logic for each
        http = requests.adapters.HTTPAdapter(max_retries=http_retries)
        https = requests.adapters.HTTPAdapter(max_retries=https_retries)

        # Replace the session’s original adapters
        s.mount('http://', http)
        s.mount('https://', https)

        r = s.get(S3path)

        log.debug(f'GET form S3: {S3path}')

        if (r.status_code >= 200) and (r.status_code < 300):
            log.debug('File downloaded from {}'.format(S3path))
            return r

        else:
            log.error(
                'File download from {} ended with error (error code: {}). Response: {}'.format(S3path, r.status_code,
                                                                                               r.text))
            return r



import logging
import sys

from .configuration import *
# Set logger with proper format for backend parsing to status page
from .constants import LOGGER_NAME

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
log = logging.Logger(LOGGER_NAME)

logger_handler = logging.StreamHandler()
logger_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(logger_handler)
config = configuration.Configuration()

# Import the methods
from .DatalakeHandler import *
from .TaskHandler import *
from .StatusHandler import *
from .Timeseries import *



import csv
import os
import json

from datetime import datetime
from . import LOGGER_NAME
from .HTTPCalller import HTTPCaller
from .DatalakeHandler import *

log = logging.getLogger(LOGGER_NAME)


class Timeseries:
    def __init__(self):
        self.caller = HTTPCaller()
        self.uploader = DatalakeHandler()

    def get_symbols(self, search_for: str=None, group_name: str = None, _size=1000, _from=0):
        '''
        Returns a list of the symbols available on a given group or symbols within search
        '''
        formatted_search = None
        formatted_group = None

        assert (group_name not in ["", None]) or (search_for not in ["", None]), 'Please specify either group_name or search_for parameter of get_symbols() method'

        # format group name and search for url - changes spaces to %20 etc.
        if group_name not in ["", None]:
            formatted_group = urllib.parse.quote(group_name)
        if search_for not in ["", None]:
            formatted_search = urllib.parse.quote(search_for)


        if formatted_search is not None and formatted_group is not None:
            # both search and group passed
            query_url = '/ts/symbol/search?size={}&from={}&query={}*%26groupName.keyword%3D{}'.format(_size, _from, formatted_search, formatted_group)
        elif formatted_search is not None and formatted_group is None:
            # search passed
            query_url = '/ts/symbol/search?size={}&from={}&query={}'.format(_size, _from, formatted_search)
        else:
            # group passed
            query_url = '/ts/symbol/search?size={}&from={}&query=groupName.keyword%3D{}'.format(_size, _from, formatted_search, formatted_group)
            

        # GET Request
        r = self.caller.get(path=query_url)

        return json.load(io.BytesIO(r.content))

    def retrieve_data_as_csv(self, symbols: dict, columns, group_name: str,
                             file_name: str = None, start_date: str = None, end_date: str = None,
                             timezone: str = None,
                             metadata: bool = False, allow_wildcard: bool = False, NCSV: bool = True,
                             corrections : str = 'false', delta: bool = False):

        '''
        The method retrieves the data from NG Timeseries based on symbols and columns.
        Optionally start and end dates can be specified. If no dates specified, all the available data are taken.

        :param file_name: the name of file locally to save data to, with the proper extension passed (for example data.csv)
        :param symbols: symbols to identify data in Timeseries. Must dictionary {"symbolName": "symbolValue", ...}
        :param columns: name of column or list of columns to retrieve
        :param group_name: name of the group
        :param start_date: start date of data in the format YYYY-MM-DD. If time is included, it must follow ISO format:
        YYYY-MM-DDTHH:mm:ss (note the 'T' letter separating the date and time part)
        :param end_date: end date of the data in the format YYYY-MM-DD. If time is included, it must follow ISO format:
        YYYY-MM-DDTHH:mm:ss (note the 'T' letter separating the date and time part)

        :param: timezone: timezone as string used for the timestamps in the Date column. Continent/City, for example Europe/London
                        if None, it returns the date with the default timezone of the user, set in the application
        :param: corrections: How to handle corrections in time series. Must be either "true", "false" or "only"
        :param: delta: delta will filter timeseries by insertion date instead of the actual timestamp of data
        :param: metadata: boolean to return or not the metadata in the file, default as False
        :param: allow_wildcard: boolean to allow or not the use of wildcards '*' in the keys or columns
        :param: corrections: How to handle corrections in time series. Must be either "true", "false", "history" or "only"
        :param: delta: delta will filter timeseries by insertion date instead of the actual timestamp of data

        :return: saves file with data
        '''

        # start_date is optional: if not provided, the data are taken from the very beginning
        if start_date is not None:
            # date only
            try:
                datetime.strptime(start_date, "%Y-%m-%d")
                start_date = start_date + "T00:00:00"
            except:
                # if cannot parse to date only, it may contain time
                try:
                    datetime.strptime(start_date, "%Y-%m-%dT%H:%M:%S")
                except:
                    raise Exception(f'Wrong format of date is passed for start date.'
                                    f'The accepted datetime formats are: YYYY-MM-DD or YYYY-MM-DDTHH:mm:ss. '
                                    f'E.g., 2021-02-05 or 2021-02-05T12:30:00')

        # start_date is optional: if not provided, the data are taken from the very beginning
        if end_date is not None:
            # date only
            try:
                datetime.strptime(end_date, "%Y-%m-%d")
                # till the end of day
                end_date = end_date + "T23:59:59"
            except:
                # if cannot parse to date only, it may contain time
                try:
                    datetime.strptime(end_date, "%Y-%m-%dT%H:%M:%S")
                except:
                    raise Exception(f'Wrong format of date is passed for end date.'
                                    f'The accepted datetime formats are: YYYY-MM-DD or YYYY-MM-DDTHH:mm:ss.'
                                    f'E.g., 2021-02-05 or 2021-02-05T12:30:00'
                                    )

        # Cast columns argument as list
        if not isinstance(columns, list):
            columns = [columns]


        # Payload Body of the REQUEST
        body = {"startDate": start_date,
                "endDate": end_date,
                "keys": [
                    {"symbols": symbols,
                    "groupName": group_name,
                    "columns": columns,
                    "pattern": allow_wildcard
                     }
                ],
                "corrections": corrections,
                "delta": delta,
                "formatType": "NCSV" if NCSV else 'PANDAS'
                }

        # add extra fields in the payload if needed
        if timezone is not None:
            # Check the format for the timezone
            assert '/' in timezone, 'Wrongly formatted timezone argument. Must be for example Europe/London'
 
            # return dates in wanted timezone
            body["timeZone"] = timezone

        # return metadata
        body['metadata'] = metadata

        # Check that if a wild_card was passed, the patter was set to True
        if ('*' in body['keys'][0]['symbols'].values()) and (allow_wildcard is False):
            raise ValueError('If passing a wildcard in the symbols, allow_wildcard must be True')


        # Summary
        log.debug(f'Retrieving data for symbols: {symbols} '
                  f'\nand columns: {columns}'
                  f'\nfrom date: {start_date}'
                  f'\nto date: {end_date}')

        # Request
        r = self.caller.post(path='/ts', payload=body)
        log.debug(f'Response: {r.text}')

        try:

            # Get file as NCSV API format
            if NCSV:
                # return raw bytes directly
                fileIO = io.BytesIO(r.content)

                if file_name is None:
                    # return the raw Bytes (streamed)
                    return fileIO

                else:
                    # Check that the extension is in the file, else add it as csv by default
                    file_name = file_name if '.' in file_name else file_name + '.csv'

                    # Save the downloaded data on the disk
                    with open(file_name, "w") as text_file:
                        text_file.write(r.text)

            # Get file as PANDAS API format
            else:
                # return the JSON for the Horizontal format
                file_json = json.load(io.BytesIO(r.content))['data']

                if file_name is None:
                    # return raw JSON format of the data
                    return file_json

                else:
                    # Check that the extension is in the file, else add it as csv by default
                    file_name = file_name if '.' in file_name else file_name + '.csv'

                    # Save the downloaded data on the disk
                    with open(file_name, "w") as file_to_write:
                        json.dump(file_json, file_to_write)

        except :
            # Get unique list of symbols
            raw_list = self.get_symbols(group_name)['items']
            symb_list = set([item for sublist in [list(d['symbols'].keys()) for d in raw_list] for item in sublist])

            # Raise error with symbols list
            raise ValueError('Wrong Symbols or Columns passed to the query . It should match the data structure on the group: {}'.format(symb_list))

        log.debug('Data are saved to file: {}'.format(file_name))

        return

    def upload_data(self, file_name, symbols_columns, date_column, group_name):
        '''
        The methods uploads data to Timeseries by re-arranging the columns order and uploading the new file as NCSV.
        :param file_name: csv file to be upload to Timeseries
        :param symbols_columns: columns to be used as symbols
        :param date_column: date column in the file; is renamed to DateTime as expected by NCSV
        :param group_name: group name for the NCSV file to upload
        :return: id of the NCSV file uploaded
        '''

        # get all column names form csv file
        with open(file_name) as csv_file:
            csv_reader = csv.DictReader(csv_file)
            dict_from_csv = dict(list(csv_reader)[0])
            all_cols = list(dict_from_csv.keys())

        log.debug(f'Original file {file_name} contains columns: {all_cols}')

        assert len(all_cols) != 0, "No columns in csv"
        assert all([x in all_cols for x in symbols_columns]), f"Symbols are not in columns list {symbols_columns}"
        assert date_column in all_cols, f"No column {date_column} in the list"

        # create new order of columns => symbols, date column, value columns
        key = symbols_columns + ['DateTime']
        value_columns = [x for x in all_cols if x not in key and x != date_column]
        final_cols = key + value_columns

        log.debug(f'Columns will be re-ordered to NCSV format: {final_cols}')

        # to keep file in the same place where the original file is
        parent_folder = os.path.dirname(file_name)
        temp_file = os.path.join(parent_folder, 'temp.csv')

        # write reordered csv to the temporary file
        with open(file_name, 'r') as infile, open(temp_file, 'a') as outfile:
            reader = csv.DictReader(infile)
            writer = csv.writer(outfile)

            # write header
            writer.writerow(final_cols)
            for row in reader:
                # change date_column to DateTime
                row['DateTime'] = row.pop(date_column)

                # prepare new line as per new columns order
                line = [row[x] for x in final_cols]
                writer.writerow(line)
        log.debug(f'File with re-ordered column is saved as {temp_file}')

        # upload file as NCSV to the specified group
        fid = self.uploader.upload_file(file=temp_file, group_name=group_name, file_type='NCSV')
        log.debug(f'File is uploaded as NCSV: {fid}')

        os.remove(temp_file)
        log.debug(f'Temporary file {temp_file} is removed.')

        return fid



import os
from .constants import API_ENDPOINT


class Configuration:
    def __init__(self):
        self.reload()

    def reload(self):
        # authentication process
        self._token = os.getenv('NG_API_AUTHTOKEN')
        self._login = os.getenv('LOGIN')
        self._password = os.getenv('PASSWORD')
        # system env variables
        self._job_id = os.environ.get('JOBID')
        self._pipeline_id = os.environ.get('PIPELINE_ID')
        self._eid = os.environ.get('EID')
        self._endpoint = os.environ.get('NG_API_ENDPOINT', API_ENDPOINT)
        self._component_name = os.environ.get('NG_COMPONENT_NAME')
        self._group_name = os.environ.get('NG_STATUS_GROUP_NAME', '')

        # Authentication from Token or login/password
        if self._token is None or self._token == '':
            if self._login is None or self._login == '':
                raise Exception("No LOGIN found. Set the environment variable with the LOGIN")
            if self._password is None or self._password == '':
                raise Exception("No PASSWORD found. Set the environment variable with the PASSWORD")

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def job_id(self):
        return self._job_id

    @property
    def eid(self):
        return self._eid

    @property
    def component_name(self):
        return self._component_name

    @property
    def group_name(self):
        return self._group_name

    @property
    def pipeline_id(self):
        return self._pipeline_id

    @property
    def token(self):
        return self._token

    @property
    def login(self):
        return self._login

    @property
    def password(self):
        return self._password



import base64
from . import *
import logging

log = logging.getLogger(LOGGER_NAME)


class Authenticator:
    def __init__(self):
        self.token = config.token
        self.login = config.login
        self.password = config.password

    def get_token(self):
        if self.token is not None:
            log.debug('Authentication with Web Token ...')
            return {'Authorization': f'Bearer {self.token}'}

        else:
            if self.login is not None and self.password is not None:
                log.debug('Authentication with Credentials ...')
                credentials = self.login + ":" + self.password
                message_bytes = credentials.encode('ascii')
                base64_enc = base64.b64encode(message_bytes).decode('UTF-8')

                d = {'Authorization': f"Basic {base64_enc}"}
                return d

            else:
                log.error('No login or password provided, neither token can be read from environment variable.')
                raise Exception('Authentication issue : no credentials found')



import os

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name=os.getenv('LIBRARY_NAME'),
    version=os.getenv('LIBRARY_VERSION'),
    author=os.getenv('LIBRARY_AUTHOR'),
    author_email=os.getenv('LIBRARY_AUTHOR_EMAIL'),
    description=os.getenv('SETUP_DESCRIPTION'),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=os.getenv('COMPANY_HOMEPAGE'),
    packages=setuptools.find_packages(),
    classifiers=["Programming Language :: Python :: 3",
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: OS Independent"
                 ],
    install_requires = [
        'requests>=2.27.1',
        'urllib3>=1.26.9'
    ],
    python_requires='>=3.6',
)
